from numpy import (
    array as _np_array, 
    exp as _np_exp,
    unique, linspace, invert, flip, transpose, concatenate, sign, zeros, 
    min as _np_min, max as _np_max, equal, where, logical_or, logical_and, all, newaxis)
from math import inf as _math_inf, pi as _math_pi, acos as _math_acos, sin as _math_sin, log2 as _math_log2
from shapely.geometry import (
    Polygon as _shapely_Polygon,
    MultiPoint as _shapely_MultiPoint,
    MultiPolygon as _shapely_MultiPolygon,
    GeometryCollection as _shapely_GeometryCollection
)
from shapely.ops import unary_union as _shapely_unary_union
from pandas import DataFrame as _pd_DataFrame
from concave_hull import concave_hull
from matplotlib import pyplot as plt
from matplotlib.patches import (Rectangle as _plt_Rectangle, Polygon as _plt_Polygon, Circle as _plt_Circle)
from .illustrations.plot_utils import plot_polygon
from .utils.general import count_polygon_interiors
from .utils.intersections import circle_line_segment_intersection
from aabpl.utils.distances_to_cell import (get_cells_relevant_for_disk_by_type, get_cells_by_lvl_overlapped_by_cell_buffer)
from aabpl.testing.test_performance import time_func_perf
# TODO make individual function for each hull type?
# TODO make clever use of buffered cells s.t. polygon check is only performed in edge cases

# Define the Polygon and the cutting line

def get_full_cell_poly(
        x_steps:list=[],
        y_steps:list=[],
):
    return _shapely_Polygon([
                (x_steps[0], y_steps[0]),
                (x_steps[-1], y_steps[0]),
                (x_steps[-1], y_steps[-1]),
                (x_steps[0], y_steps[-1]),
            ])

def split_poly_along_grid(
        poly, 
        cell_to_poly:dict={},
        lvls_to_store:list=True, 
        min_steps:int=1,
        lvl:int=None,
        x_steps:list=[],
        y_steps:list=[],
        store_full_poly:bool=True,
        row:int=0,
        col:int=0
        ) -> dict:
    """
    Split (Multi-)Polygon along grid. Store splitted polygons in dict that is returned.
    Dict contains only cells that contain a part of polygon
    """
    if lvl is None:
        # crop initial shape to grid extent
        if hasattr(poly, 'intersection'):
            poly=poly.intersection(get_full_cell_poly(x_steps, y_steps)) 
    #
        lvl = min(int((-_math_log2(len(x_steps)-1)//1)), int((-_math_log2(len(y_steps)-1)//1)))

    if type(poly) is bool and poly==True:
        pass
    elif type(poly) in [_shapely_MultiPolygon, _shapely_GeometryCollection]:
        for p in poly.geoms:
            # print("type p",type(p),lvl)
            split_poly_along_grid(
                poly=p, 
                cell_to_poly=cell_to_poly,
                lvls_to_store=lvls_to_store,
                min_steps=min_steps, 
                lvl=lvl, 
                x_steps=x_steps, 
                y_steps=y_steps, 
                store_full_poly=store_full_poly,
                row=row, 
                col=col
                )
        return cell_to_poly
    elif type(poly) is not _shapely_Polygon:
        return cell_to_poly
    elif poly.is_empty:
        return cell_to_poly
    # elif poly.area >= spacing**2*2**-(2*lvl): split_full_cells_along_grid()
    # print("len x_steps",len(x_steps), "len y_steps", len(y_steps), "split_x",split_x,"split_y",split_y,"lvl",lvl,"row",row,"col",col)
    minx, maxx = x_steps[0], x_steps[-1]
    miny, maxy = y_steps[0], y_steps[-1]
    
    lvl_x = int((-_math_log2(len(x_steps)-1)//1))
    lvl_y = int((-_math_log2(len(y_steps)-1)//1))
    split_x = int(len(x_steps)>min_steps+1 and lvl_x <= lvl_y and lvl_x <= lvl)
    split_y = int(len(y_steps)>min_steps+1 and lvl_y <= lvl_x and lvl_y <= lvl)
    # x_mid_id = (len(x_steps)//2)
    # y_mid_id = (len(y_steps)//2)
    if split_x:
        x_mid_id = 2**(-int((-_math_log2(len(x_steps)-1)//1))-1)
        midx = x_steps[x_mid_id]
    if split_y:
        y_mid_id = 2**(-int((-_math_log2(len(y_steps)-1)//1))-1)
        midy = y_steps[y_mid_id]
    
    # also dont split_x if len(y_steps)>2*len(x_steps) is 
    
    if split_x or split_y:
        if split_x:
            x_starts, x_stops = [0,x_mid_id], [x_mid_id+1, len(x_steps)]
            x_mins, x_maxs = [minx, midx], [midx, maxx]
        else:
            x_starts, x_stops = [0], [len(x_steps)]
            x_mins, x_maxs = [minx], [maxx]
        if split_y:
            y_starts, y_stops = [0,y_mid_id], [y_mid_id+1, len(y_steps)]
            y_mins, y_maxs = [miny, midy], [midy, maxy]
        else:
            y_starts, y_stops = [0], [len(y_steps)]
            y_mins, y_maxs = [miny], [maxy]

        for x_start, x_stop, x_min, x_max in zip(x_starts, x_stops, x_mins, x_maxs):
            for y_start, y_stop, y_min, y_max in zip(y_starts, y_stops, y_mins, y_maxs):
                if type(poly) is bool and poly==True:
                    intersection_poly = poly
                else:
                    cell_poly = get_full_cell_poly([x_min,x_max], [y_min, y_max])
                    intersection_poly = cell_poly.intersection(poly)
                    if cell_poly.area <= intersection_poly.area:
                        intersection_poly = True
                if not store_full_poly and type(intersection_poly) is bool and intersection_poly is True:
                    cell_to_poly[(lvl+1,(row+y_start,col+x_start))] = [intersection_poly]
                    continue
                
                split_poly_along_grid(
                    poly=intersection_poly,
                    cell_to_poly=cell_to_poly,
                    min_steps=min_steps,
                    lvl=lvl+1, 
                    store_full_poly=store_full_poly,
                    lvls_to_store=lvls_to_store,
                    x_steps=x_steps[x_start:x_stop],
                    y_steps=y_steps[y_start:y_stop],
                    row=row+y_start,
                    col=col+x_start
                )
                

        lvl_i = min(lvl_x, lvl_y, lvl)
        if ((type(lvls_to_store)==bool and lvls_to_store) or 
            (type(lvls_to_store) in [list,tuple] and lvl_i in lvls_to_store)):
            if store_full_poly and type(poly) is bool and poly is True:
                poly = get_full_cell_poly(x_steps, y_steps)
            if (lvl,(row,col)) not in cell_to_poly:
                cell_to_poly[(lvl_i,(row,col))] = [poly]
            else: 
                cell_to_poly[(lvl_i,(row,col))].append(poly)

    # elif min(lvl_x,lvl_y) <= lvl:
        # pass

    else:
        # for lvl_i in [lvl]:
        if store_full_poly and type(poly) is bool and poly is True:
            poly = get_full_cell_poly(x_steps, y_steps)

        for lvl_i in range(lvl, max([lvl_x, lvl_y, lvl])+1):
            if (lvl_i,(row,col)) not in cell_to_poly:
                cell_to_poly[(lvl_i,(row,col))] = [poly]
            else: 
                cell_to_poly[(lvl_i,(row,col))].append(poly)
            #
        #
        if min(lvl_x,lvl_y) > lvl:
            split_poly_along_grid(
                    poly=poly,
                    cell_to_poly=cell_to_poly,
                    min_steps=min_steps,
                    lvl=lvl+1, 
                    store_full_poly=store_full_poly,
                    lvls_to_store=lvls_to_store,
                    x_steps=x_steps,
                    y_steps=y_steps,
                    row=row,
                    col=col
                )
    #
    
    return cell_to_poly

@time_func_perf
def process_sample_poly_to_grid(
        grid,
        rel_tol = 0.00000,

):
    grid.sample_col_min = sample_col_min = int((grid.sample_area.bounds[0] - grid.total_bounds.xmin) // grid.spacing)
    grid.sample_row_min = sample_row_min = int((grid.sample_area.bounds[1] - grid.total_bounds.ymin) // grid.spacing)
    grid.sample_col_max = sample_col_max = int((grid.sample_area.bounds[2] - grid.total_bounds.xmin) // grid.spacing)
    grid.sample_row_max = sample_row_max = int((grid.sample_area.bounds[3] - grid.total_bounds.ymin) // grid.spacing)
    
    col_min = min(grid.col_ids)
    row_min = min(grid.row_ids)
    col_max = max(grid.col_ids)
    row_max = max(grid.row_ids)
    grid.sample_x_steps = _np_array(
        [grid.total_bounds.xmin + pad*grid.spacing for pad in range(grid.sample_col_min-min(grid.col_ids), 0)] +  
        list(grid.x_steps) + 
        [grid.total_bounds.xmax + pad*grid.spacing for pad in range(1, grid.sample_col_max - max(grid.col_ids) + 1)]
    )
    grid.sample_y_steps = _np_array(
        [grid.total_bounds.ymin + pad*grid.spacing for pad in range(grid.sample_row_min-min(grid.row_ids), 0)] +  
        list(grid.y_steps) + 
        [grid.total_bounds.ymax + pad*grid.spacing for pad in range(1, grid.sample_row_max - max(grid.row_ids) + 1)]
    )
    # grid.sample_x_steps = _np_array(
    #     [grid.total_bounds.xmin - (col_min-col)*grid.spacing for col in range(sample_col_min, col_min)] +  
    #     list(grid.x_steps) + 
    #     [grid.total_bounds.xmax + (col-col_max)*grid.spacing for col in range(max(grid.col_ids),col_max+1)]
    # )
    # grid.sample_y_steps = _np_array(
    #     [grid.total_bounds.ymin - (row_min-row)*grid.spacing for row in range(sample_row_min, row_min)] +  
    #     list(grid.y_steps) + 
    #     [grid.total_bounds.ymax + (row-row_max)*grid.spacing for row in range(max(grid.row_ids),row_max+1)]
    # )
    grid.sample_col_ids = range(col_min,col_max+1)
    grid.sample_row_ids = range(row_min,row_max+1)
    grid.sample_grid_bounds = [
        grid.total_bounds.xmin + sample_col_min * grid.spacing,
        grid.total_bounds.ymin + sample_row_min * grid.spacing,
        grid.total_bounds.xmin + (sample_col_max+1) * grid.spacing,
        grid.total_bounds.ymin + (sample_row_max+1) * grid.spacing,
    ]
    
    

    lvls_to_store = True or set([0,grid.ref_lvl])

    cell_to_poly = split_poly_along_grid(
        poly=grid.sample_area,
        cell_to_poly={},
        lvls_to_store=lvls_to_store,
        min_steps=1, 
        # lvl=None, 
        store_full_poly=False, # change to False. 
        x_steps=grid.sample_x_steps, 
        y_steps=grid.sample_y_steps, 
        row=grid.sample_row_min,
        col=grid.sample_col_min,
    )
    #split_poly_along_grid + split into subcells? 
    # unary union of polygons
    
    used_poly_lvls = sorted(set([lvl for lvl,(row,col) in cell_to_poly if lvls_to_store==True or lvl in lvls_to_store]))
    min_lvl = int(min(used_poly_lvls))
    max_lvl = int(max(used_poly_lvls))
    abs_tol = rel_tol * (grid.spacing*2**-max_lvl)**2
    # method to obtain poky for fully valid cells from lvl, row, col
    def get_fully_valid_poly(
            lvl:int,
            row,
            col, 
            sample_x_steps=grid.sample_x_steps, 
            sample_y_steps=grid.sample_y_steps, 
            max_lvl:int=max_lvl, 
            sample_row_min=grid.sample_row_min, 
            sample_col_min=grid.sample_col_min) -> dict:
        """
        Creates fully valid poly for given cell by lvl, row, col
        """
        return get_full_cell_poly(
            x_steps=sample_x_steps[col-sample_col_min:col-sample_col_min+2**(max_lvl-lvl)+1],
            y_steps=sample_y_steps[row-sample_row_min:row-sample_row_min+2**(max_lvl-lvl)+1],
        )
    grid.get_fully_valid_poly = get_fully_valid_poly

    # create unary union for cells with multiple polys
    cell_to_poly = {
        k:_shapely_unary_union(v) if len(v)!=1 else v[0] 
        for k,v in sorted([(k,v) for k,v in cell_to_poly.items()])
        }
    # extract fully valid and partly valid cells
    cells_fully_valid  ={
        (lvl, (row,col)) for (lvl, (row,col)), poly in cell_to_poly.items() 
        if lvl in used_poly_lvls and (
        (type(poly) is bool and poly is True) or
        poly.area >= (grid.spacing*2**-lvl)**2 - abs_tol
        )}
    cells_partly_valid_max_lvl = {
        (lvl, (row,col)) for (lvl, (row,col)), poly in cell_to_poly.items() 
        if lvl == max_lvl and not type(poly) is bool and
        (grid.spacing*2**-lvl)**2 - abs_tol > poly.area > abs_tol
        } 
    #

    # function to get cell_to_poly with fully valid polys instead of True for fully valid cells
    def get_cell_to_poly(grid=grid, bind_to_grid:bool=True):
        """
        get cell_to_poly dict with fully valid polys instead of True for fully valid cells
        
        :param grid: Grid
        :param bind_to_grid: whether to bind the result to the grid object
        """
        cell_to_poly = {
            (lvl,(row,col)): 
                get_fully_valid_poly(lvl,row,col) if type(v) is bool and v is True else v 
            for (lvl,(row,col)),v in grid.cell_to_poly.items()
            }
        if bind_to_grid:
            grid.cell_to_poly = cell_to_poly
        return cell_to_poly
    grid.get_cell_to_poly = get_cell_to_poly
    #

    # ensure that no subcell is part of sample_cells if parent cell is part of sample_cells
    
    
    # if store_full_poly==False then this will only remove cells that are now considered fully valid after removing precision issues.
    sample_cells = cells_fully_valid.union(cells_partly_valid_max_lvl) 
    # work from most detailed level to most aggregated
    for lvl in range(min_lvl, max_lvl+1)[::-1]:
        cell_parent = (min_lvl,(0,0))
        for row,col in [(row,col) for lvl_i,(row,col) in sample_cells if lvl_i==lvl]:
            cell_parent_included = False

            row_rounded, col_rounded = int(round(row)), int(round(col))
            row_remainder, col_remainder = (row+.5)%1, (col+.5)%1
            # for lvl_j in range(min_lvl, lvl)[::-1]:
            #     if lvl_j < 0:
            #         row_lvl_lookup, col_lvl_lookup = row_rounded//(2**lvl_j), col_rounded//(2**lvl_j)
            #     elif lvl_j == 0:
            #         row_lvl_lookup, col_lvl_lookup = row_rounded, col_rounded
            #     else:
            #         row_remainder = (row_remainder//2**-lvl //2 *2+1)*2**-lvl
            #         col_remainder = (col_remainder//2**-lvl //2 *2+1)*2**-lvl
            #         row_lvl_lookup, col_lvl_lookup = row_rounded+row_remainder-.5, col_rounded+col_remainder-.5
            #     cell_parent = (lvl_j, (row_lvl_lookup, col_lvl_lookup))
            #     cell_parent_in_fully_valid = cell_parent in cells_fully_valid
            #     cell_parent_in_partly_valid = cell_parent in cells_partly_valid
            #     cell_parent_included = cell_parent in sample_cells
            #     if cell_parent_included:
            #         break
            if lvl-1 < 0:
                # find next row, col in parent lvl which means the next smaller value out of all possible parent cell values
                parent_lvl_row_ids = [x for x in range(grid.sample_row_min, grid.sample_row_max+1) if (x-grid.sample_row_min)%(2**-(lvl-1))==0][::-1]
                parent_lvl_col_ids = [x for x in range(grid.sample_col_min, grid.sample_col_max+1) if (x-grid.sample_col_min)%(2**-(lvl-1))==0][::-1]
                row_lvl_lookup  = next((x for x in parent_lvl_row_ids if x<=row_rounded), parent_lvl_row_ids[0])
                col_lvl_lookup  = next((x for x in parent_lvl_col_ids if x<=col_rounded), parent_lvl_col_ids[0])               
                # row_lvl_lookup, col_lvl_lookup = int(row_rounded//(2**(lvl-1))), int(col_rounded//(2**(lvl-1)))
            elif (lvl-1) == 0:
                row_lvl_lookup, col_lvl_lookup = row_rounded, col_rounded
            else:
                row_remainder = (row_remainder//2**-lvl //2 *2+1)*2**-lvl
                col_remainder = (col_remainder//2**-lvl //2 *2+1)*2**-lvl
                row_lvl_lookup, col_lvl_lookup = row_rounded+row_remainder-.5, col_rounded+col_remainder-.5
            cell_parent = ((lvl-1), (row_lvl_lookup, col_lvl_lookup))
            cell_parent_in_fully_valid = cell_parent in cells_fully_valid
            cell_parent_in_partly_valid = cell_parent in cells_partly_valid_max_lvl
            cell_parent_included = cell_parent in sample_cells
         
            if cell_parent_in_fully_valid:
                if (lvl, (row, col)) in cells_fully_valid:
                    cells_fully_valid.remove((lvl, (row, col)))
                if (lvl, (row, col)) in cells_partly_valid_max_lvl:
                    cells_partly_valid_max_lvl.remove((lvl, (row, col)), cell_parent)
            
            # if cell_parent_in_partly_valid:
            #     cells_partly_valid.remove(cell_parent)

            # else keep cell


    grid.cell_to_poly = cell_to_poly
    grid.cell_to_poly_partly_valid = {cell:poly for cell, poly in cell_to_poly.items() if cell in cells_partly_valid_max_lvl}
    grid.cells_fully_valid = cells_fully_valid
    grid.cells_partly_valid_max_lvl = set(cells_partly_valid_max_lvl)
    
    # if area_complexity>500:
    #     print('WARNING: The Polygon defining the valid area is complex (n edges=' +
    #           str(area_complexity) +'). Consider simplifying it with a higher tolerance: (p=p.simplify(tolerance=...)) ')

    

def infer_sample_area_from_pts(
        pts:_pd_DataFrame=None,
        grid=None,
        hull_type:str=['buff_non_empty_cells', 'buff_cells_min_pts', 'buff_pts', 'concave','convex','bounding_box','grid'][0],
        concavity:float=1,
        buffer:float=None,
        tolerance:float=None,
        x:str='lon',
        y:str='lat',
        min_pts_to_sample_cell:int=1,
        plot_sample_area:dict=None,
) -> _shapely_Polygon:
    """Creates and returns a polygon containing all points which can be used to draw random points within

    Args:
    -------
    pts (pandas.DataFrame):
        DataFrame of points for which a search for other points within the specified radius shall be performed
    hull_type (str):
        Must be one of ['concave','convex','bounding_box','grid']. 
            - 'buff_non_empty_cells': each non-empty cell plus buffer around them
            - 'buff_cells_min_pts': each cell with at least min_pts_to_sample_cell points plus buffer around them
            - 'buff_pts': a buffer will be drawn around each point. Warning: extremely be slow for large number of points
            - 'concave': a concave hull will be drawn around points. 
            - 'convex': a convex hull will be drawn around points.
            - 'bounding_box': a bounding box around points will be drawn
            - 'grid': a box covering full grid will be drawn
    concavity (float):
        will only be used when hull_type=='concave'. Value must be in (0,Inf]. Small values results in "very concave"(=fuzzy) hull. Inf results in convex hull (default=1)
    buffer (float):
        Size of the buffer that shall be applied on the hull. If None then it will be set equal to radius (r) (default=None)
    tolerance (float):
        Tolerance>=0 used to simplify geometry using Douglas-Peucker specifying maximum allowed geometry displacement. Chosing a parameter that is too small might result in performance issues. (default=None)
    x (str):
        column name of x-coordinate (=longtitude) in pts_source (default='lon')
    y (str):
        column name of y-coordinate (=lattitude) in pts_source (default='lat')
    min_pts_to_sample_cell (int):
        will only be used when hull_type=='buff_cells_min_pts'. Minimum number of points that need to be present in a cell so that the cell is included in the sample area (default=0)
    
    Returns:
    -------
    sample_poly (shapely.geometry.Polygon):
        a grid covering all points (custom class containing 
    """
    # To-Do maybe add minumum observations per cell to be kept
    if tolerance is None:
        tolerance = buffer

    area_missing_from_hull = 1
    
    if hull_type=='buffer':
        print("sample_area hull_type 'buffer' deprectated. Use 'buff_pts' instead.")
        hull_type='buff_pts'
    elif hull_type == 'buffered_cells':
        print("sample_area hull_type 'buffered_cells' deprectated. Use 'buff_non_empty_cells' or 'buff_cells_min_pts' instead.")
        hull_type='buff_cells_min_pts'
    
    if hull_type == 'bounding_box':
    
        min_x, min_y = pts[[x,y]].values.min(axis=0)
        max_x, max_y = pts[[x,y]].values.max(axis=0)
        hull_coordinates = [
            (min_x,min_y),
            (max_x,min_y),
            (max_x,max_y),
            (min_x,max_y),
            ]
        
    elif hull_type == 'grid':
        if grid is None:
            raise ValueError('In order to use the grid bounds as valid area, a grid needs to be supplied as function input: infer_sample_area_from_pts(grid=...)')
        hull_coordinates = [
            (grid.total_bounds.xmin,grid.total_bounds.ymin),
            (grid.total_bounds.xmax,grid.total_bounds.ymin),
            (grid.total_bounds.xmax,grid.total_bounds.ymax),
            (grid.total_bounds.xmin,grid.total_bounds.ymax),
            ]
    
    elif hull_type in ['concave', 'convex']:
    
        hull_coordinates = concave_hull(
            points=pts[[x,y]].values,
            concavity=concavity if hull_type=='concave' else _math_inf,
        )
    elif hull_type == 'buff_pts':
        if len(pts)>10000:
            print("WARNING: creating a buffer around each point might cause long computation times for "+str(len(pts))+" points. Consider using hull_type='buff_non_empty_cells' or hull_type='concave' as more efficient method.")
        
        q=max(1,-(-2*buffer/tolerance)//1)
        sample_poly = _shapely_MultiPoint(pts[[x,y]].values).buffer(distance=buffer, quad_segs=q).simplify(tolerance)
        # don't simplify this shape
        area_missing_from_hull = 0

    elif hull_type in ['buff_non_empty_cells', 'buff_cells_min_pts']:
        
        grid_xmin = grid.total_bounds.xmin
        grid_ymin = grid.total_bounds.ymin
        grid_xmax = grid.total_bounds.xmax
        grid_ymax = grid.total_bounds.ymax
        if min_pts_to_sample_cell == 0 and hull_type != 'buff_non_empty_cells':
            sample_poly = _shapely_Polygon([
                (grid_xmin, grid_ymin),
                (grid_xmax, grid_ymin),
                (grid_xmax, grid_ymax),
                (grid_xmin, grid_ymax),
                ])
            
            # THIS SECTION IS REDUNDANT BECAUSE IT IS HANDLED IN process_sample_poly_to_grid
            # (contained_cells, overlapped_cells
            # ) = get_cells_by_lvl_overlapped_by_cell_buffer(grid_spacing=grid.spacing, r=buffer, nest_depth=grid.ref_lvl)
            # cells_fully_valid = set(grid.get_all_ids())
            # cells_partly_valid = set()

            # contained_cells = set([(lvl,(abs(row),abs(col))) for lvl,(row,col) in contained_cells])
            # overlapped_cells = set([(lvl,(abs(row),abs(col))) for lvl,(row,col) in overlapped_cells])
            # min_row_id = min(grid.row_ids)
            # max_row_id = max(grid.row_ids)
            # min_col_id = min(grid.col_ids)
            # max_col_id = max(grid.col_ids)
            # for row_i in grid.row_ids:
            #     for lvl,(row_j, col_j) in contained_cells:
            #         cells_fully_valid.add((lvl,(row_i-row_j,min_col_id-col_j)))
            #         cells_fully_valid.add((lvl,(row_i-row_j,max_col_id+col_j)))
            #         cells_fully_valid.add((lvl,(row_i+row_j,min_col_id-col_j)))
            #         cells_fully_valid.add((lvl,(row_i+row_j,max_col_id+col_j)))
            #     for lvl,(row_j, col_j) in overlapped_cells:
            #         cells_partly_valid.add((lvl,(row_i-row_j,min_col_id-col_j)))
            #         cells_partly_valid.add((lvl,(row_i-row_j,max_col_id+col_j)))
            #         cells_partly_valid.add((lvl,(row_i+row_j,min_col_id-col_j)))
            #         cells_partly_valid.add((lvl,(row_i+row_j,max_col_id+col_j)))
            
            # for col_i in grid.col_ids:
            #     for lvl,(row_j, col_j) in contained_cells:
            #         cells_fully_valid.add((lvl,(min_row_id-row_j,col_i-col_j)))
            #         cells_fully_valid.add((lvl,(min_row_id-row_j,col_i+col_j)))
            #         cells_fully_valid.add((lvl,(max_row_id+row_j,col_i-col_j)))
            #         cells_fully_valid.add((lvl,(max_row_id+row_j,col_i+col_j)))
            #     for lvl,(row_j, col_j) in overlapped_cells:
            #         cells_partly_valid.add((lvl,(min_row_id-row_j,col_i-col_j)))
            #         cells_partly_valid.add((lvl,(min_row_id-row_j,col_i+col_j)))
            #         cells_partly_valid.add((lvl,(max_row_id+row_j,col_i-col_j)))
            #         cells_partly_valid.add((lvl,(max_row_id+row_j,col_i+col_j)))
      
            # cells_partly_valid.difference_update(cells_fully_valid)

            # grid.cells_fully_valid = cells_fully_valid
            # grid.cells_partly_valid = cells_partly_valid

        else:
            if hull_type == 'buff_non_empty_cells':
                if min_pts_to_sample_cell != 1:
                    print("sample_area hull_type 'buff_non_empty_cells' used together with min_pts_to_sample_cell != 1. min_pts_to_sample_cell will be set = 1. Use 'buff_cells_min_pts' to specify different value.")
                min_pts_to_sample_cell = 1
            if grid.ref_lvl == 0:
                id_to_pt_ids = [id for id, val in grid.id_to_pt_ids.items() if len(val)>=min_pts_to_sample_cell]
                spacing = grid.spacing
                # print("rc__",min(grid.row_ids), max(grid.row_ids), min(grid.col_ids), max(grid.col_ids))
                # print("xy__",min([x for x,y in id_to_pt_ids]), max([x for x,y in id_to_pt_ids]), min([y for x,y in id_to_pt_ids]), max([y for x,y in id_to_pt_ids]))
                
            elif grid.ref_lvl > 0:#TODO this will cause error because of lvl, (row,col)
                spacing = spacing / (2**grid.ref_lvl)
                id_to_pt_ids = [id for id, val in grid.id_to_pt_ids_by_lvl[grid.ref_lvl].items() if len(val)>=min_pts_to_sample_cell]
                pass
            else:
                spacing = spacing * (2**(-1*grid.ref_lvl))
                id_to_pt_ids = [id for id, val in grid.id_to_pt_ids_by_lvl[grid.ref_lvl].items() if len(val)>=min_pts_to_sample_cell]
                pass
            polygons = [[grid.get_cell_poly(row,col)] for row,col in id_to_pt_ids]
            
            # THIS SECTION IS REDUNDANT BECAUSE IT IS HANDLED IN process_sample_poly_to_grid
            # buffer/spacing
            # maybe create one np array arround with coords of buffered cell around (0,0) and then add centroids to it 
            # (contained_cells, overlapped_cells
            # ) = get_cells_by_lvl_overlapped_by_cell_buffer(grid_spacing=grid.spacing, r=buffer, nest_depth=grid.ref_lvl)
            # cells_fully_valid = set([(0,(row,col)) for row,col in id_to_pt_ids])
            
            # cells_partly_valid = set()
            # for row,col in id_to_pt_ids:
            #     for lvl,(d_row,d_col) in contained_cells:
            #         tp = int if lvl==0 else float
            #         cells_fully_valid.add( (lvl, (tp(row+d_row), tp(col+d_col))) )
            #     for lvl,(d_row,d_col) in overlapped_cells:
            #         tp = int if lvl==0 else float
            #         cells_partly_valid.add( (lvl, (tp(row+d_row), tp(col+d_col))) )        
            # cells_partly_valid.difference_update(cells_fully_valid)
            # grid.cells_fully_valid = cells_fully_valid
            # grid.cells_partly_valid = cells_partly_valid

            sample_poly = _shapely_MultiPolygon(polygons).buffer(buffer, quadsegs=3)
            # don't simplify this shape
        area_missing_from_hull = 0
           
    else:
        raise ValueError("hull_type to infere sample area for random points must be in ['buff_non_empty_cells', 'buff_cells_min_pts', 'concave','convex','buff_pts', 'bounding_box', 'grid']. Value provided:",hull_type)
    #
    if area_missing_from_hull > 0:
        hull_poly = _shapely_Polygon(hull_coordinates).buffer(distance=0,quad_segs=1)
        # plot_polygon(poly=hull_poly)
        while area_missing_from_hull > 0:
            sample_poly = hull_poly
            q=max(1,-(-2*buffer/tolerance)//1)
            sample_poly = sample_poly.buffer(distance=buffer, quad_segs=q)
            
            sample_poly = sample_poly.simplify(tolerance)

            area_missing_from_hull = hull_poly.difference(
                sample_poly.intersection(hull_poly)
            ).area
            tolerance = tolerance*0.8
        #

    # plot_polygon(poly=sample_poly)
    if not plot_sample_area is None:
        fig,ax = plt.subplots(nrows=1,ncols=1,figsize=(8,8))
        ax.scatter(x=pts[x], y=pts[y], color="#51da58", s=0.3)
        plot_polygon(ax=ax, poly=sample_poly, facecolor="#06047640", edgecolor='red')
    # shoot warning if polygon is getting comple

    return sample_poly
#

def remove_invalid_area_from_sample_poly(
        sample_poly:_shapely_Polygon,
        invalid_areas:_shapely_Polygon,
):
    valid_area_poly = sample_poly.difference(invalid_areas)

    return valid_area_poly
#

def apply_invalid_area_on_grid(
        
):
    return
#

def disk_cell_intersection_area(
    disk_center_pt:_np_array,      
    row_col:tuple=(1,2),
    grid_spacing:float=0.0075,
    r:float=0.0075,
    silent = False,
    return_n_itx=False,
) -> float:
    """
    note this does not handle the case where the point lies within cell bounds

    Calculates intersection area of cell and search-circle (0,grid_spacing**2)
    Case for no intersection will be handled before (fully included or fully excluded).
    Case 1: two intersection points (more than half of square are within radius) - 3 vertices are within circle 
    Case 2: two intersection points (more than half of square are within radius) - 1 vertex is within circle
    Case 3: two intersection points (less than half of square within radius) - 0 vertices within circle (same row or col)
    Case 4: two intersection points (unclear wheter more or less than half) - 2 vertices within circle (same row or col)
    Case 5: four intersection points (more than half of circle is included) - 2 vertices within circle (same row or col)
    
    TODO: if grid_spacing/2 is greater than radius there will be weird instances 

    This can also be done already as a function of the point offset
    and is also symmetrical towards the triangle
    the intersection area only needs to be computed for those cases where excluded cells are intersected 

    """
    
    (xmin,ymin),(xmax,ymax) = (row_col[0]-.5)*grid_spacing, (row_col[1]-.5)*grid_spacing, (row_col[0]+.5)*grid_spacing, (row_col[1]+.5)*grid_spacing
    if grid_spacing is None:
        rectangle_area = (abs(xmax-xmin)*abs(ymax-ymin))
    else:
        rectangle_area = grid_spacing**2
    calculated_area = 0
    precision = 13
    if not silent:
        fig,ax=plt.subplots()
        ax.add_patch(_plt_Circle(xy=disk_center_pt, radius=r, alpha=0.4))
        ax.add_patch(_plt_Rectangle(xy=(xmin,ymin),width=(xmax-xmin),height=(ymax-ymin), alpha=0.4))
        ax.autoscale_view()
    
    vtx_coords = (
        (xmin,ymin),
        (xmax,ymin),
        (xmax,ymax),
        (xmin,ymax),
        )
    vtx_dis_to_c = (
        ((disk_center_pt[0]-vtx_coords[0][0])**2+(disk_center_pt[1]-vtx_coords[0][1])**2)**.5,
        ((disk_center_pt[0]-vtx_coords[1][0])**2+(disk_center_pt[1]-vtx_coords[1][1])**2)**.5,
        ((disk_center_pt[0]-vtx_coords[2][0])**2+(disk_center_pt[1]-vtx_coords[2][1])**2)**.5,
        ((disk_center_pt[0]-vtx_coords[3][0])**2+(disk_center_pt[1]-vtx_coords[3][1])**2)**.5
    )
    vtx_in_r = [dis<=r for dis in vtx_dis_to_c]
    
    if sum(vtx_in_r)==4:
        if return_n_itx:
            return (rectangle_area, sum(vtx_in_r)) 
        else:
            return rectangle_area
    #
    if sum(vtx_in_r)==0:
        # circular segment 
        # say vertices B and C are closest to circle center. Then circle intersects BC 
        # (twice unless only touching or not intersecting at all)
        v_closest1, v_secondclosest1 = [i for d,i in sorted([(d,i) for i,d in enumerate(vtx_dis_to_c)])][:2]
        segment1 = (vtx_coords[v_closest1], vtx_coords[v_secondclosest1])
        itx_pts = circle_line_segment_intersection(
            circle_center=disk_center_pt,
            circle_radius=r,
            pt1=segment1[0],
            pt2=segment1[1],
            full_line=False,
            precision=precision,
        )
        if len(itx_pts)<2:
            if return_n_itx:
                return (0., sum(vtx_in_r))
            else:
                return 0.
        
        itx_pt1, itx_pt2 = itx_pts
        #             
    elif sum(vtx_in_r)==1:
        # circular segment + triangle
        # say vertex B is closest to circle center then circle intersect AB and BC
        v_closest1 = vtx_dis_to_c.index(min(vtx_dis_to_c))
        segment1 = (vtx_coords[(v_closest1-1)%4], vtx_coords[v_closest1])
        segment2 = (vtx_coords[v_closest1], vtx_coords[(v_closest1+1)%4])
        itx_pt1 = circle_line_segment_intersection(
            circle_center=disk_center_pt,
            circle_radius=r,
            pt1=segment1[0],
            pt2=segment1[1],
            full_line=False,
            precision=precision,
        )
        itx_pt2 = circle_line_segment_intersection(
            circle_center=disk_center_pt,
            circle_radius=r,
            pt1=segment2[0],
            pt2=segment2[1],
            full_line=False,
            precision=precision,
        )
        if len(itx_pt1) != 1 or len(itx_pt2) != 1:
            raise ValueError("Unexpected number of intersections",itx_pt1, itx_pt2)
        itx_pt1, itx_pt2 = itx_pt1[0], itx_pt2[0]
        triangle = 1/2 * abs(itx_pt1[0]-itx_pt2[0])*abs(itx_pt1[1]-itx_pt2[1])
        calculated_area += triangle    
    elif sum(vtx_in_r)==2:
        # circular segement + triangle + triangle
        # say vertex B and C are closest to circle center then circle intersect AB and CD
        v_closest1, v_secondclosest1 = [i for d,i in sorted([(d,i) for i,d in enumerate(vtx_dis_to_c)])][:2]
        offset = 1 if (v_closest1-1)%4 != v_secondclosest1 else -1 
        segment1 = (vtx_coords[(v_closest1-offset)%4], vtx_coords[v_closest1])
        segment2 = (vtx_coords[v_secondclosest1], vtx_coords[(v_secondclosest1+offset)%4])
        itx_pt1 = circle_line_segment_intersection(
            circle_center=disk_center_pt,
            circle_radius=r,
            pt1=segment1[0],
            pt2=segment1[1],
            full_line=False,
            precision=precision,
        )
        itx_pt2 = circle_line_segment_intersection(
            circle_center=disk_center_pt,
            circle_radius=r,
            pt1=segment2[0],
            pt2=segment2[1],
            full_line=False,
            precision=precision,
        )
        if len(itx_pt1) != 1 or len(itx_pt2) != 1:
            raise ValueError("Unexpected number of intersections",itx_pt1, itx_pt2)
        itx_pt1, itx_pt2 = itx_pt1[0], itx_pt2[0]
        # match: which intersection point is aligned to which vertex?
        # align_closesty_to_pt1y = (
        #     0 if vtx_coords[v_closest1][0]==itx_pt1[0] else # x closest = x pt1 
        #     1 if vtx_coords[v_closest1][1]==itx_pt1[1] else # y closest = y pt1
        #     -2 if vtx_coords[v_closest1][0]==itx_pt2[0] else # x closest = x pt2
        #     -1 #if vtx_coords[v_closest1][0]==itx_pt1[0] # y closest = y pt2
        # )
        # # triangle 1: closest1,secondclosest1,pt2
        # pt_A, pt_B = (itx_pt1, itx_pt2) if align_closesty_to_pt1y in [0,1] else (itx_pt2, itx_pt1)
        # x_is_aligned = int(align_closesty_to_pt1y%2==0)
        # # if x is aligned take y difference
        # if vtx_coords[v_closest1][x_is_aligned]!=vtx_coords[v_secondclosest1][x_is_aligned]:
        #     print("x_is_aligned",x_is_aligned,"closest1",vtx_coords[v_closest1],"secondclosest1",vtx_coords[v_secondclosest1])
        # triangles = 1/2 * grid_spacing * (
        #     abs(vtx_coords[v_closest1][x_is_aligned]-pt_B[x_is_aligned]) +
        #     abs(vtx_coords[v_secondclosest1][x_is_aligned]-pt_A[x_is_aligned])
        # ) 
       
        x_is_aligned = 1 if vtx_coords[v_closest1][0]==itx_pt1[0] or vtx_coords[v_closest1][0]==itx_pt2[0] else 0
        triangles = 1/2 * grid_spacing * (
            abs(vtx_coords[v_closest1][x_is_aligned]-itx_pt1[x_is_aligned]) +
            abs(vtx_coords[v_closest1][x_is_aligned]-itx_pt2[x_is_aligned])
        )

        calculated_area += triangles

        # triangle 2: closest1,pt2,pt1
        # triangle1 = 1/2 * abs(
        #     vtx_coords[v_closest1][(align_closesty_to_pt1y+1)%2]-itx_pt2[(align_closesty_to_pt1y+1)%2]) * abs(
        #         vtx_coords[v_closest1][(align_closesty_to_pt1y+0)%2]-itx_pt2[(align_closesty_to_pt1y+0)%2])
        # # triangle 2: closest1,pt2,pt1
        # triangle2 = 1/2 * abs(
        #     vtx_coords[v_closest1][(align_closesty_to_pt1y+1)%2]-itx_pt1[(align_closesty_to_pt1y+1)%2]) * abs(
        #     vtx_coords[v_closest1][(align_closesty_to_pt1y+0)%2]-itx_pt2[(align_closesty_to_pt1y+0)%2]
        # )
        # # if triangle1==0.0 or triangle2==0.0:
        # #     print("triangle1",triangle1, "triangle2", triangle2)
        # # else:
        # #     print("both,,,")
        # calculated_area += triangle1 + triangle2
    elif sum(vtx_in_r)==3:
        # rectangle - triangle + circular segement
        # say vertex B is most distant from circle center: then circle intersects line segments AB and BC.
        v_farthest1 = vtx_dis_to_c.index(max(vtx_dis_to_c))
        segment1 = (vtx_coords[(v_farthest1-1)%4], vtx_coords[v_farthest1])
        segment2 = (vtx_coords[v_farthest1], vtx_coords[(v_farthest1+1)%4])
        
        itx_pt1 = circle_line_segment_intersection(
            circle_center=disk_center_pt,
            circle_radius=r,
            pt1=segment1[0],
            pt2=segment1[1],
            full_line=False,
            precision=precision,
        )
        itx_pt2 = circle_line_segment_intersection(
            circle_center=disk_center_pt,
            circle_radius=r,
            pt1=segment2[0],
            pt2=segment2[1],
            full_line=False,
            precision=precision,
        )
        itx_pt1, itx_pt2 = itx_pt1[0], itx_pt2[0]
        triangle = 1/2 * abs(itx_pt1[0]-itx_pt2[0])*abs(itx_pt1[1]-itx_pt2[1])
        calculated_area = rectangle_area - triangle 
    
    # calcualte area of circle segement
    len_pt0_pt1 = ((itx_pt1[0]-itx_pt2[0])**2+(itx_pt1[1]-itx_pt2[1])**2)**.5
    angle_rad = _math_pi-2*_math_acos(((len_pt0_pt1/2)/r))
    circle_segment_area = 0.5 * r**2 * (angle_rad - _math_sin(angle_rad))
    calculated_area += circle_segment_area

    # return  (calculated_area, sum(vtx_in_r))
    if return_n_itx:
        return min(calculated_area, rectangle_area),sum(vtx_in_r)
    else:
        return min(calculated_area, rectangle_area)

#

def disk_cell_intersection_estimate(
    disk_center_pt:_np_array,      
    centroid:tuple=(0,0),
    grid_spacing:float=0.0075,
    r:float=0.0075,
    a:float=0,
    b:float=0,
        
):
    d = ((disk_center_pt[0]-centroid[0])**2+(disk_center_pt[0]-centroid[0])**2)**.5
    d_r = d/r
    s_r = grid_spacing/r

    # b and c are the same for all points
    b = 1 / (0.70628102 + _np_exp(0.57266908 * (s_r - 2))) # b
    c = 1 / (-0.21443453 + _np_exp(0.76899004 * (s_r - 2))) # c
    a = 1 - 1 / (1.0 + b * _np_exp(-c * (d_r - 1)))
    area = a * grid_spacing**2
    return area

def calculated_valid_area_around_pts(
        pts,
        grid,
        r:float,
        invalid_grid_cells,
        x:str='lon',
        y:str='lat',
        invalid_area_geometry=None,
        return_percentage:bool=True,
):
    """Calculates valid
    return_percentage if True returns share from [0,1] otherwise returns area in units of projection (meters)



    """
    full_circle = 2*_math_pi*r**2
    valid_area = full_circle
    # sort points by cell - cell_region - xy 
    # for cell: common contained cells sum invalid area
    # check if overlapped cells have an invalid area
    # if not jump to next cell
    # for cell-cell-region sum common contained area
    # check if overlapped cell have an invalid area
    # if not jump to next cell-region
    # for point check overlapped cells for invalid area



from warnings import simplefilter 
from pandas.errors import PerformanceWarning as _pd_PerfromanceWarning
simplefilter(action="ignore", category=_pd_PerfromanceWarning)
simplefilter(action='ignore', category=FutureWarning)
from pandas import DataFrame as _pd_DataFrame
from numpy import array as _np_array, nan as _np_nan
from .random_distribution import get_distribution_for_random_points
from .valid_area import infer_sample_area_from_pts, remove_invalid_area_from_sample_poly, process_sample_poly_to_grid
from aabpl.testing.test_performance import time_func_perf
from aabpl.radius_search.radius_search_class import DiskSearch
from aabpl.radius_search.grid_class import Grid
from aabpl.illustrations.plot_pt_vars import create_plots_for_vars
from aabpl.illustrations.distribution_plot import create_distribution_plot
from aabpl.utils.general import count_polygon_edges,find_column_name
from aabpl.utils.crs_transformation import convert_MultiPolygon_crs, convert_coords_to_local_crs, convert_pts_to_crs, convert_wgs_to_utm
from shapely.geometry import (Polygon as _shapely_Polygon, MultiPolygon as _shapely_MultiPolygon)

def check_kwargs(
        pts:_pd_DataFrame,
        crs:str,
        sample_area_crs:str,
        r:float,
        c:list=[],
        aggregation_method:str='sum',
        x:str='lon',
        y:str='lat',
        row_name:str='id_y',
        col_name:str='id_x',
        sum_suffix:str=None,
        pts_target:_pd_DataFrame=None,
        x_tgt:str=None,
        y_tgt:str=None,
        row_name_tgt:str=None,
        col_name_tgt:str=None,
        grid:Grid=None,
        nest_depth:int=None,
        proj_crs:str='auto',
        silent:bool=None,
):
    """
    check shared keyword arguments and apply defaults
    """
    # locals() TODO use locals to make this take in only locals
    if type(row_name) != str:
        raise TypeError('`row_name` must be of type str. Instead provided of type',type(row_name),row_name)
    if type(col_name) != str:
        raise TypeError('`col_name` must be of type str. Instead provided of type',type(col_name),col_name)
    if row_name_tgt is None:
        row_name_tgt = row_name
    elif type(row_name_tgt) != str:
        raise TypeError('`row_name_tgt` must be of type str. Instead provided of type',type(row_name_tgt),row_name_tgt)
    if col_name_tgt is None:
        col_name_tgt = col_name
    elif type(col_name_tgt) != str:
        raise TypeError('`col_name_tgt` must be of type str. Instead provided of type',type(col_name_tgt),col_name_tgt)
    if type(pts) != _pd_DataFrame:
        raise TypeError('`pts` must be a pandas.DataFrame or None. Instead provided of type',type(pts))
    if type(x) != str:
        raise TypeError('`x` must be of type str. Instead provided of type',type(x),x)
    if type(y) != str:
        raise TypeError('`x` must be of type str. Instead provided of type',type(y),y)
    if not x in pts.columns:
        raise ValueError('`x` (x-coord column name) must be in columns of pts')
    if not y in pts.columns:
        raise ValueError('`y` (y-coord column name) must be in columns of pts')
    if not type(sum_suffix) is str:
        if not sum_suffix is None:
            sum_suffix = str(sum_suffix)
        else:
            r_suffix = int(r) if r%1==0 or len(str(int(r))) > 5 else round(r,6-len(str(int(r))))
            sum_suffix = '_' + str(r_suffix)+'m'
    if x_tgt is None:
        x_tgt = x
    if y_tgt is None:
        y_tgt = y
    same_target = pts_target is None or pts is pts_target
    if pts_target is None:
        pts_target = pts
    else:
        if type(pts_target) != _pd_DataFrame:
            raise TypeError('`pts_target` must be a pandas.DataFrame or None. Instead provided of type',type(pts_target))
    if type(c) == str:
        c = [c]
    else:
        if c is None or len(c)==0:
            print("Warning: No columns specified for aggregation - will simply count number of points within radius.")
            aggregation_method = 'count'
        try:
            if any([type(column)!=str for column in c]):
                raise TypeError
        except:
            raise TypeError('`c` must be either a string of single column name or a list of column name strings')
    if any([not column in pts_target.columns for column in c]):
        raise ValueError('not all columns(',c,') are in columns of search target pts_target(',pts.columns,')')
    if not x_tgt in pts_target.columns:
        raise ValueError('`x_tgt` (x-coord column name) must be in columns of pts_target')
    if not y_tgt in pts_target.columns:
        raise ValueError('`y_tgt` (y-coord column name) must be in columns of pts_target')
    if sample_area_crs is None:
        sample_area_crs = crs
    if proj_crs == 'auto': 
        x_center = (min([pts[x].min(), pts_target[x_tgt].min()])+max([pts[x].max(), pts_target[x_tgt].max()]))/2
        y_center = (min([pts[x].min(), pts_target[x_tgt].min()])+max([pts[x].max(), pts_target[x_tgt].max()]))/2
        local_crs = 'EPSG:'+str(convert_wgs_to_utm(x_center, y_center))
    else:
        local_crs = proj_crs
    if crs != local_crs:
        x,y,local_crs = convert_pts_to_crs(pts=pts, x=x, y=y, initial_crs=crs, target_crs=proj_crs)
        if not same_target:
            x_tgt,y_tgt,local_crs = convert_pts_to_crs(pts=pts_target, x=x_tgt, y=y_tgt, initial_crs=crs, target_crs=proj_crs)
        else:
            x_tgt,y_tgt = x,y
    
    # OVERWRITE DEFAULTS
    if grid is None:
        if not nest_depth is None and int(nest_depth) != nest_depth:
            raise TypeError('`nest_depth` must be either of type int or None. Instead:', nest_depth, "of type "+str(type(nest_depth))+' was provided.')
        elif not nest_depth is None:
            nest_depth = int(nest_depth)
        grid = create_auto_grid_for_radius_search(
            pts_source=pts,
            initial_crs=local_crs,
            local_crs=local_crs,
            data_crs=crs,
            r=r,
            nest_depth=nest_depth,
            x=x,
            y=y,
            pts_target=pts_target,
            x_tgt=x_tgt,
            y_tgt=y_tgt,
            silent=silent,
        )
    elif type(grid) != Grid:
        raise TypeError('`grid` must be either of type Grid or None. Instead:', grid, "of type "+str(type(grid))+' was provided.')


    return (pts, local_crs,  sample_area_crs, c, x, y, sum_suffix, pts_target, x_tgt, y_tgt, row_name_tgt, col_name_tgt, grid, aggregation_method)
#


@time_func_perf
def handle_sample_area_input(
    pts:_pd_DataFrame,
    r:float,
    sample_area='buffered_cells',
    sample_area_crs=None, 
    local_crs:str=None,
    x:str='lon',
    y:str='lat',
    grid:Grid=None,
    min_pts_to_sample_cell:int=0,
    no_plot:bool=True
):
    if type(sample_area)==bool and sample_area==False:
        return _shapely_Polygon([
            (grid.total_bounds.xmin, grid.total_bounds.ymin),
            (grid.total_bounds.xmax, grid.total_bounds.ymin),
            (grid.total_bounds.xmax, grid.total_bounds.ymax),
            (grid.total_bounds.xmin, grid.total_bounds.ymax)
        ])

    if sample_area is None:
        sample_area = 'grid'
    if type(sample_area) == str:
        if no_plot:
            print("Creating sample area with method '"+sample_area+"' and buffer=tolerance="+str(r)+". Use 'grid.sample_area' to inspect.")
        sample_area = infer_sample_area_from_pts(
            pts=pts,
            grid=grid,
            x=x,
            y=y,
            hull_type=sample_area,
            buffer=r,
            min_pts_to_sample_cell=min_pts_to_sample_cell,
            plot_sample_area=None,
        )
        # sample_area = remove_invalid_area_from_sample_poly(sample_area, invalid_areas=_shapely_Polygon([]))

    elif type(sample_area) in [_shapely_Polygon, _shapely_MultiPolygon]:
        sample_area = convert_MultiPolygon_crs(multipoly=sample_area, initial_crs=sample_area_crs,target_crs=local_crs)
    else:
        raise ValueError('sample_area must parameter most be one of ["str","Poylgon","MultiPolygon"] instead of type', type(sample_area))
    
    
    
    return sample_area

# TODO remove cell_region from kwargs
@time_func_perf
def create_auto_grid_for_radius_search(
    pts_source:_pd_DataFrame,
    initial_crs:str,
    local_crs:str,
    data_crs:str,
    r:float,
    nest_depth:int=None,
    x:str='lon',
    y:str='lat',
    pts_target:_pd_DataFrame=None,
    x_tgt:str=None,
    y_tgt:str=None,
    silent:bool=None,
):
    """
    Returns a Grid that covers all points and will 
    - can be used to represent clusters
    - and is leverage for performance gains of radius search 

    Args:
    -------
    pts_source (pandas.DataFrame):
        DataFrame of points for which a search for other points within the specified radius shall be performed
    crs (str):
        crs of coordinates, e.g. 'EPSG:4326'
    r (float):
        radius within which other points shall be found in meters 
    x (str):
        column name of x-coordinate (=longtitude) in pts_source (default='lon')
    y (str):
        column name of y-coordinate (=lattitude) in pts_source (default='lat')
    pts_target (pandas.DataFrame):
        DataFrame of points that shall be found/aggregated when searching within radius of points from pts_source. If None specified its assumed to be the same as pts_source. (default=None)
    x_tgt (str):
        column name of x-coordinate (=longtitude) in pts_target. If None its assumed to be same as x (default=None)
    y_tgt (str):
        column name of y-coordinate (=lattitude) in pts_target. If None its assumed to be same as y (default=None)
    silent (bool):
        Whether information on progress shall be printed to console (default=False)
    
    Returns:
    -------
    grid (aabl.Grid):
        a grid covering all points (custom class containing 
    """

    if pts_target is None:
        xmin = pts_source[x].min()
        xmax = pts_source[x].max()
        ymin = pts_source[y].min()
        ymax = pts_source[y].max()
    else:
        if y_tgt is None:
            y_tgt = y
        if x_tgt is None:
            x_tgt = x
        xmin = min([pts_source[x].min(), pts_target[x_tgt].min()])
        xmax = max([pts_source[x].max(), pts_target[x_tgt].max()])
        ymin = min([pts_source[y].min(), pts_target[y_tgt].min()])
        ymax = max([pts_source[y].max(), pts_target[y_tgt].max()])
    
    return Grid(
            xmin=xmin,
            xmax=xmax,
            ymin=ymin,
            ymax=ymax,
            initial_crs=initial_crs,
            local_crs=local_crs,
            set_fixed_spacing=r/3, # TODO don t set fixed spacing but
            r=r,
            n_pts_src=len(pts_source),
            n_pts_tgt=len(pts_target),
            nest_depth=nest_depth,
            data_crs=data_crs,
            silent=silent,
        )
#

@time_func_perf
def radius_search(
    pts:_pd_DataFrame,
    crs:str,
    r:float,
    c:list=[],
    exclude_pt_itself:bool=True,
    weight_valid_area:str=None,
    sample_area=False,
    sample_area_crs:str=None,
    include_boundary:bool=False,
    aggregation_method:str=['sum','count','mean'][0],
    x:str='lon',
    y:str='lat',
    row_name:str='id_y',
    col_name:str='id_x',
    sum_suffix:str='_r_sum', 
    pts_target:_pd_DataFrame=None,
    x_tgt:str=None,
    y_tgt:str=None,
    row_name_tgt:str=None,
    col_name_tgt:str=None,
    grid:Grid=None,
    nest_depth:int=None,
    proj_crs:str='auto',
    plot_pt_disk:dict=None,
    plot_cell_reg_assign:dict=None,
    plot_offset_checks:dict=None,
    plot_offset_regions:dict=None,
    plot_offset_raster:dict=None,
    silent:bool=None,
):
    """
    For all points in DataFrame it searches for all other points (potentially of another DataFrame) within the specified radius and aggregate the values for specified column(s)
    The result will be appended to DataFrame.

    Args:
    -------
    pts (pandas.DataFrame):
        DataFrame of points for which a search for other points within the specified radius shall be performed
    crs (str):
        crs of coordinates, e.g. 'EPSG:4326'
    r (float):
        radius within which other points shall be found in meters 
    c (str or list):
        column name or list of column name(s) in DataFrame for which data within search radius shall be aggregated. If None provided it will simply count the points within the radius. Column name must be in pts(DataFrame) unless a different search target is specified - then columns must exist in pts_target.
    exclude_pt_itself (bool):
        whether the sums within search radius point shall exlclude the point data itself (default=True)
    sample_area (shapely.geometry.Polygon | shapely.geometry.MultiPolygon | str):
        Specifies the area in which random points can be drawn.
        Either geometry is supplied directly (ensure that it uses the same projection in meters that the rest of the algorithm does).
        Or a string can be supplied that will be passed on into infer_sample_area_from_pts to create the sample area:
            - 'buffered_cells': each non-empty cell plus buffer around them
            - 'concave': a concave hull will be drawn around points. 
            - 'convex': a convex hull will be drawn around points.
            - 'buffer': points will only be drawn within buffer around points. WARNING: Very slow if many points in df.
            - 'bounding_box': a bounding box around points will be drawn
            - 'grid' or None: a box covering full grid will be drawn
        However using the function infer_sample_area_from_pts directly is recommended as it give you more control over its parameters.
        (default='buffered_cells')
    sample_area_crs (str):
        crs of the sample_area polygon. Ignored if sample_area param is not a (Multi-)Polygon. When None it will assume same crs as pts. (Default=None) 
    weight_valid_area (str):
        if set to 'estimate' or 'precise' the radius aggregate will be weighted inversely by the share of area of valid cells within search radius. 'precise' is very slow, 'estimate' has MSE of 5% of cell area. (default=None)
    x (str):
        column name of x-coordinate (=longtitude) in pts (default='lon')
    y (str):
        column name of y-coordinate (=lattitude) in pts (default='lat')
    row_name (str):
        name for column that will be appended to pts indicating grid cell row (default='id_y')
    col_name (str):
        name for column that will be appended to pts indicating grid cell column (default='id_x')
    sum_suffix (str):
        suffix used for new column(s) creating by aggregating data of columns , 
    pts_target (pandas.DataFrame):
        DataFrame of points that shall be found/aggregated when searching within radius of points from pts_source. If None specified its assumed to be the same as pts_source. (default=None)
    x_tgt (str):
        column name of x-coordinate (=longtitude) in pts_target. If None its assumed to be same as x (default=None)
    y_tgt (str):
        column name of y-coordinate (=lattitude) in pts_target. If None its assumed to be same as y (default=None)
    row_name_tgt (str):
        name for column that will be appended to pts indicating grid cell row. If None its assumed to be same as x (default=None)
    col_name_tgt (str):
        name for column that will be appended to pts indicating grid cell column (default='id_x')
    proj_crs (crs):
        crs projection into which pts for search (source and target) shall be mapped. If 'auto' local crs will be determined automatically. If None no reprojection will be performed (default='auto')
    grid (aabpl.Grid):
        grid of custom class containing points. If None it will automatically one (default=None)
    include_boundary (bool):
        FEATURE NOT YET IMPLEMENTED. Define whether points that are at the distance of exactly the radius shall be considered within (Default=False)
    plot_pt_disk (dict):
        Only needed for development. Dictionary with kwargs to create plot for example radius search. If None no plot will be created (default=None)
    plot_cell_reg_assign (dict):
        Only needed for development. Dictionary with kwargs to create plot for assginments of points to cell offset regions. If None no plot will be created (default=None)
    plot_offset_checks (dict):
        Only needed for development. Dictionary with kwargs to create plot for checks creating offset regions. If None no plot will be created (default=None)
    plot_offset_regions (dict):
        Only needed for development. Dictionary with kwargs to create plot for offset regions. If None no plot will be created (default=None)
    plot_offset_raster (dict):
        Only needed for development. Dictionary with kwargs to create plot for raster for offset regions. If None no plot will be created (default=None)
    silent (bool):
        Whether information on progress shall be printed to console (default=False)
    
    Returns:
    -------
    grid (aabl.Grid):
        a grid covering all points (custom class containing  
    
          
    Examples:
    -------
    from aabpl.main import radius_search
    from pandas import read_csv
    pts = read_csv('C:/path/to/file.txt',sep=',',header=None)
    pts.columns = ["eid", "employment", "industry", "lat","lon","moved"]
    grid = radius_search(pts,crs="EPSG:4326",r=750,columns=['employment'])
    grid.plot_vars(filename='employoment_750m')
    """
    init_sort = find_column_name('initial_sort', existing_columns=pts.columns)
    pts[init_sort] = range(len(pts))

    (pts, local_crs, sample_area_crs, c, x, y, sum_suffix, pts_target, x_tgt, y_tgt, row_name_tgt, col_name_tgt, grid, aggregation_method
     ) = check_kwargs(
            pts=pts, crs=crs, sample_area_crs=sample_area_crs, r=r, c=c, x=x, y=y, row_name=row_name,
            col_name=col_name, sum_suffix=sum_suffix, pts_target=pts_target, x_tgt=x_tgt, y_tgt=y_tgt,
            row_name_tgt=row_name_tgt, col_name_tgt=col_name_tgt, grid=grid, nest_depth=nest_depth, 
            proj_crs=proj_crs, silent=silent,
    )

    if aggregation_method in ['count','mean']:
        count_helper_col = find_column_name('count','_helper_col', existing_columns=pts_target.columns)
        c.append(count_helper_col)
        pts_target[count_helper_col] = 1
   
    # initialize disk_search
    grid.search = DiskSearch(
        grid=grid,
        r=r,
        nest_depth=nest_depth,
        exclude_pt_itself=exclude_pt_itself,
        weight_valid_area=weight_valid_area,
        include_boundary=include_boundary
    )
    # for i in range(10000):
    #     for j in range(40000):
    #         2243.8**.1324549

    # prepare target points data
    grid.search.set_target(
        pts=pts_target,
        c=c,
        x=x_tgt,
        y=y_tgt,
        row_name=row_name_tgt,
        col_name=col_name_tgt,
        silent=silent,
    )
    
    # prepare source points data
    grid.search.set_source(
        pts=pts,
        c=c,
        x=x,
        y=y,
        row_name=row_name,
        col_name=col_name,
        sum_suffix=sum_suffix,
        plot_cell_reg_assign=plot_cell_reg_assign,
        plot_offset_checks=plot_offset_checks,
        plot_offset_regions=plot_offset_regions,
        plot_offset_raster=plot_offset_raster,
        silent=silent,
    )
    
    # in case sums shall be weighted by sample area
    grid.sample_area = handle_sample_area_input(
        pts=pts,r=r,sample_area=sample_area,
        sample_area_crs=sample_area_crs,local_crs=local_crs,x=x,y=y,
        grid=grid, min_pts_to_sample_cell=0)
    process_sample_poly_to_grid(grid)
    disk_sums_for_pts = grid.search.perform_search(silent=False if silent is None else silent,plot_pt_disk=plot_pt_disk)
    
    if aggregation_method in ['mean']:
        n_rs = 1 # TODO for distance bands... later
        radius_count_cols = disk_sums_for_pts.columns[-n_rs:]
        radius_count_col = radius_count_cols[0] # TODO for distance bands... later
        for s_name in disk_sums_for_pts.columns[:-n_rs]:
            if s_name not in radius_count_cols:
                pts[s_name][pts[radius_count_col]>0] = pts[s_name][pts[radius_count_col]>0] / pts[radius_count_col][pts[radius_count_col]>0]  
                pts[s_name][pts[radius_count_col]==0] = _np_nan
        pts.drop(columns=[count_helper_col], inplace=True)
        # pts.drop(columns=[disk_sums_for_pts.columns[-1]], inplace=True)


    pts.sort_values(init_sort, inplace=True)
    pts.drop(columns=[init_sort], inplace=True)

    return grid
#

@time_func_perf
def detect_cluster_pts(
    pts:_pd_DataFrame,
    crs:str,
    r:float=0.0075,
    c:list=[],
    aggregation_method:str=['sum','count','mean'][0],
    exclude_pt_itself:bool=True,
    sample_area='buffered_cells',
    sample_area_crs:str=None,
    min_pts_to_sample_cell:int=0,
    weight_valid_area:str=None,
    k_th_percentile:float=99.5,
    n_random_points:int=int(1e5),
    random_seed:int=None,
    include_boundary:bool=False,
    x:str='lon',
    y:str='lat',
    row_name:str='id_y',
    col_name:str='id_x',
    sum_suffix:str='_750m',
    cluster_suffix:str='_cluster',
    proj_crs:str='auto',
    pts_target:_pd_DataFrame=None,
    x_tgt:str=None,
    y_tgt:str=None,
    row_name_tgt:str=None,
    col_name_tgt:str=None,
    grid:Grid=None,
    nest_depth:int=None,
    plot_distribution:dict=None,
    plot_cluster_points:dict=None,
    plot_pt_disk:dict=None,
    plot_cell_reg_assign:dict=None,
    plot_offset_checks:dict=None,
    plot_offset_regions:dict=None,
    plot_offset_raster:dict=None,
    silent:bool=None,
):
    """
    For all points in a DataFrame it searches for all other points (potentially of another DataFrame) within the specified radius and aggregate the values for specified column(s).
    It draws random the bounding box containing all points from DataFrame(s) and aggregate the values within the radius to obtain a random distribution.   
    Then all points from DataFrame which exceed the k_th_percentile of the random distribution are labeld as clustered.
    The results will be appended to DataFrame.

    Args:
    -------
    pts (pandas.DataFrame):
        DataFrame of points for which a search for other points within the specified radius shall be performed
    crs (str):
        crs of coordinates, e.g. 'EPSG:4326'
    r (float):
        radius within which other points shall be found in meters 
    c (str or list):
        column name or list of column name(s) in DataFrame for which data within search radius shall be aggregated. If None provided it will simply count the points within the radius. Column name must be in pts(DataFrame) unless a different search target is specified - then columns must exist in pts_target.
    exclude_pt_itself (bool):
        whether the sums within search radius point shall exlclude the point data itself (default=True)
    weight_valid_area (str):
        if set to 'estimate' or 'precise' the radius aggregate will be weighted inversely by the share of area of valid cells within search radius. 'precise' is very slow, 'estimate' has MSE of 5% of cell area. (default=None)
    sample_area (shapely.geometry.Polygon | shapely.geometry.MultiPolygon | str):
        Specifies the area in which random points can be drawn.
        Either geometry is supplied directly (ensure that it uses the same projection in meters that the rest of the algorithm does).
        Or a string can be supplied that will be passed on into infer_sample_area_from_pts to create the sample area:
            - 'buffered_cells': each non-empty cell plus buffer around them
            - 'concave': a concave hull will be drawn around points. 
            - 'convex': a convex hull will be drawn around points.
            - 'buffer': points will only be drawn within buffer around points. WARNING: Very slow if many points in df.
            - 'bounding_box': a bounding box around points will be drawn
            - 'grid' or None: a box covering full grid will be drawn
        However using the function infer_sample_area_from_pts directly is recommended as it give you more control over its parameters.
        (default='buffered_cells')
    sample_area_crs (str):
        crs of the sample_area polygon. Ignored if sample_area param is not a (Multi-)Polygon. When None it will assume same crs as pts. (Default=None) 
    min_pts_to_sample_cell (int):
        minimum number of points in dataset that need to be in cell s.t. random points are allowed to be drawn within it. (default=0)
    k_th_percentile (float):
        percentile of random distribution that a point needs to exceed to be classified as clustered.
    n_random_points (int):
        number of random points to be drawn to create random distribution (default=100000)
    random_seed (int):
        random seed to be applied when drawing random points to create random distribution. If None no seed will be set (default=None)
    x (str):
        column name of x-coordinate (=longtitude) in pts (default='lon')
    y (str):
        column name of y-coordinate (=lattitude) in pts (default='lat')
    row_name (str):
        name for column that will be appended to pts indicating grid cell row (default='id_y')
    col_name (str):
        name for column that will be appended to pts indicating grid cell column (default='id_x')
    sum_suffix (str):
        suffix used for new column(s) creating by aggregating data of columns , 
    pts_target (pandas.DataFrame):
        DataFrame of points that shall be found/aggregated when searching within radius of points from pts_source. If None specified its assumed to be the same as pts_source. (default=None)
    x_tgt (str):
        column name of x-coordinate (=longtitude) in pts_target. If None its assumed to be same as x (default=None)
    y_tgt (str):
        column name of y-coordinate (=lattitude) in pts_target. If None its assumed to be same as y (default=None)
    row_name_tgt (str):
        name for column that will be appended to pts indicating grid cell row. If None its assumed to be same as x (default=None)
    col_name_tgt (str):
        name for column that will be appended to pts indicating grid cell column (default='id_x')
    grid (aabpl.Grid):
        grid of custom class containing points. If None it will automatically one (default=None)
    include_boundary (bool):
        FEATURE NOT YET IMPLEMENTED. Define whether points that are at the distance of exactly the radius shall be considered within (Default=False)
    plot_distribution (dict):
        dictionary with kwargs to create plot for random distribution. If None no plot will be created (default=None)
    plot_pt_disk (dict):
        Only needed for development. Dictionary with kwargs to create plot for example radius search. If None no plot will be created (default=None)
    plot_cell_reg_assign (dict):
        Only needed for development. Dictionary with kwargs to create plot for assginments of points to cell offset regions. If None no plot will be created (default=None)
    plot_offset_checks (dict):
        Only needed for development. Dictionary with kwargs to create plot for checks creating offset regions. If None no plot will be created (default=None)
    plot_offset_regions (dict):
        Only needed for development. Dictionary with kwargs to create plot for offset regions. If None no plot will be created (default=None)
    plot_offset_raster (dict):
        Only needed for development. Dictionary with kwargs to create plot for raster for offset regions. If None no plot will be created (default=None)
    silent (bool):
        Whether information on progress shall be printed to console (default=False)
    
    Returns:
    -------
    grid (aabl.Grid):
        a grid covering all points (custom class) with cluster attributes stored to it
    """
    init_sort = find_column_name('initial_sort', existing_columns=pts.columns)
    pts[init_sort] = range(len(pts))

    (pts, local_crs, sample_area_crs, c, x, y, sum_suffix, pts_target, x_tgt, y_tgt, row_name_tgt, col_name_tgt, grid, aggregation_method
     ) = check_kwargs(
            pts=pts, crs=crs, sample_area_crs=sample_area_crs, r=r, c=c, aggregation_method=aggregation_method,
            x=x, y=y, row_name=row_name, col_name=col_name, sum_suffix=sum_suffix, 
            pts_target=pts_target, x_tgt=x_tgt, y_tgt=y_tgt, row_name_tgt=row_name_tgt, col_name_tgt=col_name_tgt, 
            grid=grid, nest_depth=nest_depth, proj_crs=proj_crs, silent=silent,
    )
    if type(k_th_percentile) not in [list,_np_array, tuple]:
        k_th_percentile = [k_th_percentile for column in c]
    elif len(k_th_percentile) < len(c):
        k_th_percentile = [k_th_percentile[i%len(k_th_percentile)] for i in range(len(c))]
    # initialize disk_search
    grid.search = DiskSearch(
        grid,
        r=r,
        nest_depth=nest_depth,
        exclude_pt_itself=exclude_pt_itself,
        weight_valid_area=weight_valid_area,
        include_boundary=include_boundary
    )

    grid.search.set_target(
        pts=pts_target,
        c=c,
        x=x_tgt,
        y=y_tgt,
        row_name=row_name_tgt,
        col_name=col_name_tgt,
        silent=silent,
    )

    #
    grid.sample_area = handle_sample_area_input(
        pts=pts, r=r, 
        sample_area=sample_area, sample_area_crs=sample_area_crs,local_crs=local_crs,x=x, y=y, grid=grid,
        min_pts_to_sample_cell=min_pts_to_sample_cell,
        no_plot=plot_distribution is None and plot_cluster_points is None)
    process_sample_poly_to_grid(grid=grid)
    
    (cluster_threshold_values, rndm_pts) = get_distribution_for_random_points(
        grid=grid,
        pts=pts,
        sample_area=grid.sample_area,
        min_pts_to_sample_cell=min_pts_to_sample_cell,
        c=c,
        x=x,
        y=y,
        row_name=row_name,
        col_name=col_name,
        sum_suffix=sum_suffix,
        n_random_points=n_random_points,
        k_th_percentile=k_th_percentile,
        random_seed=random_seed,
        silent=silent,
    )

    if not silent:
        for (colname, threshold_value, k_th_p) in zip(c, cluster_threshold_values,k_th_percentile):
            print("Threshold value for "+str(k_th_p)+"th-percentile is "+str(threshold_value)+" for "+str(colname)+" within "+str(r)+" meters.")
    

    grid.search.set_source(
        pts=pts,
        c=c,
        x=x,
        y=y,
        row_name=row_name,
        col_name=col_name,
        sum_suffix=sum_suffix,
        plot_cell_reg_assign=plot_cell_reg_assign,
        plot_offset_checks=plot_offset_checks,
        plot_offset_regions=plot_offset_regions,
        plot_offset_raster=plot_offset_raster,
        silent=silent,
    )


    disk_sums_for_pts = grid.search.perform_search(silent=silent,plot_pt_disk=plot_pt_disk)
    
    # save bool of whether pt is part of a cluster 
    for j, cname in enumerate(c):
        pts[str(cname)+str(cluster_suffix)] = disk_sums_for_pts.values[:,j]>cluster_threshold_values[j]


    if plot_distribution is not None:
        # print("disk_sums_for_random_points", disk_sums_for_random_points)
        create_distribution_plot(
            pts=pts,
            x=x,
            y=y,
            radius_sum_columns=[n+sum_suffix for n in c],
            grid=grid,
            rndm_pts=rndm_pts,
            cluster_threshold_values=cluster_threshold_values,
            k_th_percentile=k_th_percentile,
            r=r,
            plot_kwargs=plot_distribution
            )
    #

    def plot_rand_dist(
            filename:str="",
            pts=pts,
            x=x,
            y=y,
            radius_sum_columns=[n+sum_suffix for n in c],
            rndm_pts=rndm_pts,
            cluster_threshold_values=cluster_threshold_values,
            k_th_percentile=k_th_percentile,
            r=r,
            grid=grid,
            **plot_kwargs
            
    ):
        create_distribution_plot(
            filename=filename,
            plot_kwargs=plot_kwargs,
            pts=pts,
            x=x,
            y=y,
            radius_sum_columns=radius_sum_columns,
            grid=grid,
            rndm_pts=rndm_pts,
            cluster_threshold_values=cluster_threshold_values,
            k_th_percentile=k_th_percentile,
            r=r,
            )
    grid.plot.rand_dist = plot_rand_dist
    
    plot_colnames = list(c) + [n+sum_suffix for n in c] + [str(cname)+str(cluster_suffix) for cname in c]
    def plot_cluster_pts(
            self=grid,
            colnames=_np_array(plot_colnames),
            filename:str="",
            **plot_kwargs,
    ):
        return create_plots_for_vars(
            grid=self,
            colnames=colnames,
            filename=filename,
            plot_kwargs=plot_kwargs,
        )
    grid.plot.cluster_pts = plot_cluster_pts

    if plot_cluster_points is not None:
        grid.plot.cluster_pts(**plot_cluster_points)
        pass
    pts.sort_values(init_sort, inplace=True)
    pts.drop(columns=[init_sort], inplace=True)
    
    return grid
# done

def detect_cluster_cells(
    pts:_pd_DataFrame,
    crs:str,
    r:float=750,
    c:list=[],
    aggregation_method:str=['sum','count','mean'][0],
    exclude_pt_itself:bool=True,
    sample_area='buffered_cells',
    sample_area_crs:str=None,
    min_pts_to_sample_cell:int=0,
    weight_valid_area:str=None,
    k_th_percentile:float=99.5,
    n_random_points:int=int(1e5),
    random_seed:int=None,
    queen_contingency:int=1,
    rook_contingency:int=1,
    centroid_dist_threshold:float=None,
    border_dist_threshold:float=None,
    min_cluster_share_after_contingency:float=0.05,
    min_cluster_share_after_centroid_dist:float=0.00,
    min_cluster_share_after_convex:float=0.00,
    make_convex:bool=True,
    include_boundary:bool=False,
    x:str='lon',
    y:str='lat',
    row_name:str='id_y',
    col_name:str='id_x',
    sum_suffix:str='_750m',
    cluster_suffix:str='_cluster',
    proj_crs:str='auto',
    pts_target:_pd_DataFrame=None,
    x_tgt:str=None,
    y_tgt:str=None,
    row_name_tgt:str=None,
    col_name_tgt:str=None,
    grid:Grid=None,
    nest_depth:int=None,
    plot_distribution:dict=None,
    plot_cluster_points:dict=None,
    plot_pt_disk:dict=None,
    plot_cell_reg_assign:dict=None,
    plot_offset_checks:dict=None,
    plot_offset_regions:dict=None,
    plot_offset_raster:dict=None,
    silent:bool=None,
):
    """
    For all points in a DataFrame it searches for all other points (potentially of another DataFrame) within the specified radius and aggregate the values for specified column(s).
    It draws random the bounding box containing all points from DataFrame(s) and aggregate the values within the radius to obtain a random distribution.   
    Then all points from DataFrame which exceed the k_th_percentile of the random distribution are labeld as clustered.
    The results will be appended to DataFrame.

    Args:
    -------
    pts (pandas.DataFrame):
        DataFrame of points for which a search for other points within the specified radius shall be performed
    crs (str):
        crs of coordinates, e.g. 'EPSG:4326'
    r (float):
        radius within which other points shall be found in meters 
    c (str or list):
        column name or list of column name(s) in DataFrame for which data within search radius shall be aggregated. If None provided it will simply count the points within the radius. Column name must be in pts(DataFrame) unless a different search target is specified - then columns must exist in pts_target. 
    exclude_pt_itself (bool):
        whether the sums within search radius point shall exlclude the point data itself (default=True)
    weight_valid_area (str):
        if set to 'estimate' or 'precise' the radius aggregate will be weighted inversely by the share of area of valid cells within search radius. 'precise' is very slow, 'estimate' has MSE of 5% of cell area. (default=None)
    sample_area (shapely.geometry.Polygon | shapely.geometry.MultiPolygon | str):
        Specifies the area in which random points can be drawn.
        Either geometry is supplied directly (ensure that it uses the same projection in meters that the rest of the algorithm does).
        Or a string can be supplied that will be passed on into infer_sample_area_from_pts to create the sample area:
            - 'buffered_cells': each non-empty cell plus buffer around them
            - 'concave': a concave hull will be drawn around points. 
            - 'convex': a convex hull will be drawn around points.
            - 'bounding_box': a bounding box around points will be drawn
            - 'grid' or None: a box covering full grid will be drawn
        However using the function infer_sample_area_from_pts directly is recommended as it give you more control over its parameters.
        (default='buffered_cells')
    sample_area_crs (str):
        crs of the sample_area polygon. Ignored if sample_area param is not a (Multi-)Polygon. When None it will assume same crs as pts. (Default=None) 
    min_pts_to_sample_cell (int):
        minimum number of points in dataset that need to be in cell s.t. random points are allowed to be drawn within it. (default=0)
    k_th_percentile (float):
        percentile of random distribution that a point needs to exceed to be classified as clustered.
    n_random_points (int):
        number of random points to be drawn to create random distribution (default=100000)
    random_seed (int):
        random seed to be applied when drawing random points to create random distribution. If None no seed will be set (default=None)
    queen_contingency (int):
        if contigent (vertical, horizontal, diagonal) cells that are also classified as clustered shall be part of the same cluster. If set to a value>=2 then it also adds non-contingent cells that are that many steps away to the same cluster. (default=1) 
    rook_contingency (int):
        if contigent (vertical, horizontal) cells that are also classified as clustered shall be part of the same cluster. Ignored if queen_contingency is set to a higher value. If set to a value>=2 then it also adds non-contingent cells that are that many steps away to the same cluster. (default=1) 
    centroid_dist_threshold (float):
        maximum distance between centroids of clusters to be merged into a single cluster. If None clusters won't be merged based on centroid distance. (default=r*10/3)
    border_dist_threshold (float):
        maximum distance between borders of clusters to be merged into a single cluster. If None clusters won't be merged based on boundary distance (default=r*4/3)
    min_cluster_share_after_contingency (float):
        minimum share of cluster of total to not be dropped after cells are merged to clusters based on contingency
    min_cluster_share_after_centroid_dist (float):
        minimum share of cluster of total to not be dropped after clusters are merged based on centroid
    min_cluster_share_after_convex (float):
        minimum share of cluster of total to not be dropped after clusters are made convex by adding cells within its convex hull
    make_convex (bool):
        Whether all cells within the convex hull of a cluster shall be added to it (default=True)
    include_boundary (bool):
        FEATURE NOT YET IMPLEMENTED. Define whether points that are at the distance of exactly the radius shall be considered within (Default=False)
    x (str):
        column name of x-coordinate (=longtitude) in pts (default='lon')
    y (str):
        column name of y-coordinate (=lattitude) in pts (default='lat')
    row_name (str):
        name for column that will be appended to pts indicating grid cell row (default='id_y')
    col_name (str):
        name for column that will be appended to pts indicating grid cell column (default='id_x')
    sum_suffix (str):
        suffix used for new column(s) creating by aggregating data of columns , 
    pts_target (pandas.DataFrame):
        DataFrame of points that shall be found/aggregated when searching within radius of points from pts_source. If None specified its assumed to be the same as pts_source. (default=None)
    x_tgt (str):
        column name of x-coordinate (=longtitude) in pts_target. If None its assumed to be same as x (default=None)
    y_tgt (str):
        column name of y-coordinate (=lattitude) in pts_target. If None its assumed to be same as y (default=None)
    row_name_tgt (str):
        name for column that will be appended to pts indicating grid cell row. If None its assumed to be same as x (default=None)
    col_name_tgt (str):
        name for column that will be appended to pts indicating grid cell column (default='id_x')
    grid (aabpl.Grid):
        grid of custom class containing points. If None it will automatically one (default=None)
    plot_distribution (dict):
        dictionary with kwargs to create plot for random distribution. If None no plot will be created (default=None)
    plot_pt_disk (dict):
        Only needed for development. Dictionary with kwargs to create plot for example radius search. If None no plot will be created (default=None)
    plot_cell_reg_assign (dict):
        Only needed for development. Dictionary with kwargs to create plot for assginments of points to cell offset regions. If None no plot will be created (default=None)
    plot_offset_checks (dict):
        Only needed for development. Dictionary with kwargs to create plot for checks creating offset regions. If None no plot will be created (default=None)
    plot_offset_regions (dict):
        Only needed for development. Dictionary with kwargs to create plot for offset regions. If None no plot will be created (default=None)
    plot_offset_raster (dict):
        Only needed for development. Dictionary with kwargs to create plot for raster for offset regions. If None no plot will be created (default=None)
    silent (bool):
        Whether information on progress shall be printed to console (default=False)
    
    Returns:
    -------
    grid (aabl.Grid):
        a grid covering all points (custom class) with cluster attributes stored to it  
    """
    (pts, local_crs, sample_area_crs, c, x, y, sum_suffix, pts_target, x_tgt, y_tgt, row_name_tgt, col_name_tgt, grid, aggregation_method
     ) = check_kwargs(
            pts=pts, crs=crs, sample_area_crs=sample_area_crs, r=r, c=c, aggregation_method=aggregation_method,
            x=x, y=y, row_name=row_name, col_name=col_name, sum_suffix=sum_suffix, pts_target=pts_target, x_tgt=x_tgt, y_tgt=y_tgt,
            row_name_tgt=row_name_tgt, col_name_tgt=col_name_tgt, grid=grid, nest_depth=nest_depth, 
            proj_crs=proj_crs, silent=silent,
    )
    if centroid_dist_threshold is None:
        centroid_dist_threshold = r * 10/3
    if border_dist_threshold is None:
        border_dist_threshold = r * 4/3
    
    grid = detect_cluster_pts(
        pts=pts,
        crs=local_crs,
        r=r,
        c=c,
        exclude_pt_itself=exclude_pt_itself,
        weight_valid_area=weight_valid_area,
        sample_area=sample_area,
        sample_area_crs=sample_area_crs,
        min_pts_to_sample_cell=min_pts_to_sample_cell,
        k_th_percentile=k_th_percentile,
        n_random_points=n_random_points,
        random_seed=random_seed,
        grid=grid,
        nest_depth=nest_depth,
        x=x,
        y=y,
        row_name=row_name,
        col_name=col_name,
        sum_suffix=sum_suffix,
        cluster_suffix=cluster_suffix,
        proj_crs=local_crs,
        pts_target=pts_target,
        x_tgt=x_tgt,
        y_tgt=y_tgt,
        row_name_tgt=row_name_tgt,
        col_name_tgt=col_name_tgt,
        include_boundary=include_boundary,
        plot_distribution=plot_distribution,
        plot_cluster_points=plot_cluster_points,
        plot_pt_disk=plot_pt_disk,
        plot_cell_reg_assign=plot_cell_reg_assign,
        plot_offset_checks=plot_offset_checks,
        plot_offset_regions=plot_offset_regions,
        plot_offset_raster=plot_offset_raster,
        silent=silent,
    )
    
    grid.clustering.create_clusters(
        pts=pts,
        c=c,
        queen_contingency=queen_contingency,
        rook_contingency=rook_contingency,
        centroid_dist_threshold=centroid_dist_threshold,
        border_dist_threshold=border_dist_threshold,
        min_cluster_share_after_contingency=min_cluster_share_after_contingency,
        min_cluster_share_after_centroid_dist=min_cluster_share_after_centroid_dist,
        min_cluster_share_after_convex=min_cluster_share_after_convex,
        make_convex=make_convex,
        row_name=row_name,
        col_name=col_name,
        cluster_suffix=cluster_suffix,
        )
    
    return grid
#
def detect_cluster_cells_from_labeled_pts():
    pass    
def radius_sum():
    pass
def radius_count():
    pass
def radius_mean():
    pass



"""
This file is separate from ``_add_newdocs.py`` so that it can be mocked out by
our sphinx ``conf.py`` during doc builds, where we want to avoid showing
platform-dependent information.
"""
import sys
import os
from numpy._core import dtype
from numpy._core import numerictypes as _numerictypes
from numpy._core.function_base import add_newdoc

##############################################################################
#
# Documentation for concrete scalar classes
#
##############################################################################

def numeric_type_aliases(aliases):
    def type_aliases_gen():
        for alias, doc in aliases:
            try:
                alias_type = getattr(_numerictypes, alias)
            except AttributeError:
                # The set of aliases that actually exist varies between platforms
                pass
            else:
                yield (alias_type, alias, doc)
    return list(type_aliases_gen())


possible_aliases = numeric_type_aliases([
    ('int8', '8-bit signed integer (``-128`` to ``127``)'),
    ('int16', '16-bit signed integer (``-32_768`` to ``32_767``)'),
    ('int32', '32-bit signed integer (``-2_147_483_648`` to ``2_147_483_647``)'),
    ('int64', '64-bit signed integer (``-9_223_372_036_854_775_808`` to ``9_223_372_036_854_775_807``)'),
    ('intp', 'Signed integer large enough to fit pointer, compatible with C ``intptr_t``'),
    ('uint8', '8-bit unsigned integer (``0`` to ``255``)'),
    ('uint16', '16-bit unsigned integer (``0`` to ``65_535``)'),
    ('uint32', '32-bit unsigned integer (``0`` to ``4_294_967_295``)'),
    ('uint64', '64-bit unsigned integer (``0`` to ``18_446_744_073_709_551_615``)'),
    ('uintp', 'Unsigned integer large enough to fit pointer, compatible with C ``uintptr_t``'),
    ('float16', '16-bit-precision floating-point number type: sign bit, 5 bits exponent, 10 bits mantissa'),
    ('float32', '32-bit-precision floating-point number type: sign bit, 8 bits exponent, 23 bits mantissa'),
    ('float64', '64-bit precision floating-point number type: sign bit, 11 bits exponent, 52 bits mantissa'),
    ('float96', '96-bit extended-precision floating-point number type'),
    ('float128', '128-bit extended-precision floating-point number type'),
    ('complex64', 'Complex number type composed of 2 32-bit-precision floating-point numbers'),
    ('complex128', 'Complex number type composed of 2 64-bit-precision floating-point numbers'),
    ('complex192', 'Complex number type composed of 2 96-bit extended-precision floating-point numbers'),
    ('complex256', 'Complex number type composed of 2 128-bit extended-precision floating-point numbers'),
    ])


def _get_platform_and_machine():
    try:
        system, _, _, _, machine = os.uname()
    except AttributeError:
        system = sys.platform
        if system == 'win32':
            machine = os.environ.get('PROCESSOR_ARCHITEW6432', '') \
                    or os.environ.get('PROCESSOR_ARCHITECTURE', '')
        else:
            machine = 'unknown'
    return system, machine


_system, _machine = _get_platform_and_machine()
_doc_alias_string = f":Alias on this platform ({_system} {_machine}):"


def add_newdoc_for_scalar_type(obj, fixed_aliases, doc):
    # note: `:field: value` is rST syntax which renders as field lists.
    o = getattr(_numerictypes, obj)

    character_code = dtype(o).char
    canonical_name_doc = "" if obj == o.__name__ else \
                        f":Canonical name: `numpy.{obj}`\n    "
    if fixed_aliases:
        alias_doc = ''.join(f":Alias: `numpy.{alias}`\n    "
                            for alias in fixed_aliases)
    else:
        alias_doc = ''
    alias_doc += ''.join(f"{_doc_alias_string} `numpy.{alias}`: {doc}.\n    "
                         for (alias_type, alias, doc) in possible_aliases if alias_type is o)

    docstring = f"""
    {doc.strip()}

    :Character code: ``'{character_code}'``
    {canonical_name_doc}{alias_doc}
    """

    add_newdoc('numpy._core.numerictypes', obj, docstring)


_bool_docstring = (
    """
    Boolean type (True or False), stored as a byte.

    .. warning::

       The :class:`bool` type is not a subclass of the :class:`int_` type
       (the :class:`bool` is not even a number type). This is different
       than Python's default implementation of :class:`bool` as a
       sub-class of :class:`int`.
    """
)

add_newdoc_for_scalar_type('bool', [], _bool_docstring)

add_newdoc_for_scalar_type('bool_', [], _bool_docstring)

add_newdoc_for_scalar_type('byte', [],
    """
    Signed integer type, compatible with C ``char``.
    """)

add_newdoc_for_scalar_type('short', [],
    """
    Signed integer type, compatible with C ``short``.
    """)

add_newdoc_for_scalar_type('intc', [],
    """
    Signed integer type, compatible with C ``int``.
    """)

# TODO: These docs probably need an if to highlight the default rather than
#       the C-types (and be correct).
add_newdoc_for_scalar_type('int_', [],
    """
    Default signed integer type, 64bit on 64bit systems and 32bit on 32bit
    systems.
    """)

add_newdoc_for_scalar_type('longlong', [],
    """
    Signed integer type, compatible with C ``long long``.
    """)

add_newdoc_for_scalar_type('ubyte', [],
    """
    Unsigned integer type, compatible with C ``unsigned char``.
    """)

add_newdoc_for_scalar_type('ushort', [],
    """
    Unsigned integer type, compatible with C ``unsigned short``.
    """)

add_newdoc_for_scalar_type('uintc', [],
    """
    Unsigned integer type, compatible with C ``unsigned int``.
    """)

add_newdoc_for_scalar_type('uint', [],
    """
    Unsigned signed integer type, 64bit on 64bit systems and 32bit on 32bit
    systems.
    """)

add_newdoc_for_scalar_type('ulonglong', [],
    """
    Signed integer type, compatible with C ``unsigned long long``.
    """)

add_newdoc_for_scalar_type('half', [],
    """
    Half-precision floating-point number type.
    """)

add_newdoc_for_scalar_type('single', [],
    """
    Single-precision floating-point number type, compatible with C ``float``.
    """)

add_newdoc_for_scalar_type('double', [],
    """
    Double-precision floating-point number type, compatible with Python
    :class:`float` and C ``double``.
    """)

add_newdoc_for_scalar_type('longdouble', [],
    """
    Extended-precision floating-point number type, compatible with C
    ``long double`` but not necessarily with IEEE 754 quadruple-precision.
    """)

add_newdoc_for_scalar_type('csingle', [],
    """
    Complex number type composed of two single-precision floating-point
    numbers.
    """)

add_newdoc_for_scalar_type('cdouble', [],
    """
    Complex number type composed of two double-precision floating-point
    numbers, compatible with Python :class:`complex`.
    """)

add_newdoc_for_scalar_type('clongdouble', [],
    """
    Complex number type composed of two extended-precision floating-point
    numbers.
    """)

add_newdoc_for_scalar_type('object_', [],
    """
    Any Python object.
    """)

add_newdoc_for_scalar_type('str_', [],
    r"""
    A unicode string.

    This type strips trailing null codepoints.

    >>> s = np.str_("abc\x00")
    >>> s
    'abc'

    Unlike the builtin :class:`str`, this supports the
    :ref:`python:bufferobjects`, exposing its contents as UCS4:

    >>> m = memoryview(np.str_("abc"))
    >>> m.format
    '3w'
    >>> m.tobytes()
    b'a\x00\x00\x00b\x00\x00\x00c\x00\x00\x00'
    """)

add_newdoc_for_scalar_type('bytes_', [],
    r"""
    A byte string.

    When used in arrays, this type strips trailing null bytes.
    """)

add_newdoc_for_scalar_type('void', [],
    r"""
    np.void(length_or_data, /, dtype=None)

    Create a new structured or unstructured void scalar.

    Parameters
    ----------
    length_or_data : int, array-like, bytes-like, object
       One of multiple meanings (see notes).  The length or
       bytes data of an unstructured void.  Or alternatively,
       the data to be stored in the new scalar when `dtype`
       is provided.
       This can be an array-like, in which case an array may
       be returned.
    dtype : dtype, optional
       If provided the dtype of the new scalar.  This dtype must
       be "void" dtype (i.e. a structured or unstructured void,
       see also :ref:`defining-structured-types`).

       .. versionadded:: 1.24

    Notes
    -----
    For historical reasons and because void scalars can represent both
    arbitrary byte data and structured dtypes, the void constructor
    has three calling conventions:

    1. ``np.void(5)`` creates a ``dtype="V5"`` scalar filled with five
       ``\0`` bytes.  The 5 can be a Python or NumPy integer.
    2. ``np.void(b"bytes-like")`` creates a void scalar from the byte string.
       The dtype itemsize will match the byte string length, here ``"V10"``.
    3. When a ``dtype=`` is passed the call is roughly the same as an
       array creation.  However, a void scalar rather than array is returned.

    Please see the examples which show all three different conventions.

    Examples
    --------
    >>> np.void(5)
    np.void(b'\x00\x00\x00\x00\x00')
    >>> np.void(b'abcd')
    np.void(b'\x61\x62\x63\x64')
    >>> np.void((3.2, b'eggs'), dtype="d,S5")
    np.void((3.2, b'eggs'), dtype=[('f0', '<f8'), ('f1', 'S5')])
    >>> np.void(3, dtype=[('x', np.int8), ('y', np.int8)])
    np.void((3, 3), dtype=[('x', 'i1'), ('y', 'i1')])

    """)

add_newdoc_for_scalar_type('datetime64', [],
    """
    If created from a 64-bit integer, it represents an offset from
    ``1970-01-01T00:00:00``.
    If created from string, the string can be in ISO 8601 date
    or datetime format.
    
    When parsing a string to create a datetime object, if the string contains
    a trailing timezone (A 'Z' or a timezone offset), the timezone will be 
    dropped and a User Warning is given.
    
    Datetime64 objects should be considered to be UTC and therefore have an 
    offset of +0000.

    >>> np.datetime64(10, 'Y')
    numpy.datetime64('1980')
    >>> np.datetime64('1980', 'Y')
    numpy.datetime64('1980')
    >>> np.datetime64(10, 'D')
    numpy.datetime64('1970-01-11')

    See :ref:`arrays.datetime` for more information.
    """)

add_newdoc_for_scalar_type('timedelta64', [],
    """
    A timedelta stored as a 64-bit integer.

    See :ref:`arrays.datetime` for more information.
    """)

add_newdoc('numpy._core.numerictypes', "integer", ('is_integer',
    """
    integer.is_integer() -> bool

    Return ``True`` if the number is finite with integral value.

    .. versionadded:: 1.22

    Examples
    --------
    >>> np.int64(-2).is_integer()
    True
    >>> np.uint32(5).is_integer()
    True
    """))

# TODO: work out how to put this on the base class, np.floating
for float_name in ('half', 'single', 'double', 'longdouble'):
    add_newdoc('numpy._core.numerictypes', float_name, ('as_integer_ratio',
        """
        {ftype}.as_integer_ratio() -> (int, int)

        Return a pair of integers, whose ratio is exactly equal to the original
        floating point number, and with a positive denominator.
        Raise `OverflowError` on infinities and a `ValueError` on NaNs.

        >>> np.{ftype}(10.0).as_integer_ratio()
        (10, 1)
        >>> np.{ftype}(0.0).as_integer_ratio()
        (0, 1)
        >>> np.{ftype}(-.25).as_integer_ratio()
        (-1, 4)
        """.format(ftype=float_name)))

    add_newdoc('numpy._core.numerictypes', float_name, ('is_integer',
        f"""
        {float_name}.is_integer() -> bool

        Return ``True`` if the floating point number is finite with integral
        value, and ``False`` otherwise.

        .. versionadded:: 1.22

        Examples
        --------
        >>> np.{float_name}(-2.0).is_integer()
        True
        >>> np.{float_name}(3.2).is_integer()
        False
        """))

for int_name in ('int8', 'uint8', 'int16', 'uint16', 'int32', 'uint32',
        'int64', 'uint64', 'int64', 'uint64', 'int64', 'uint64'):
    # Add negative examples for signed cases by checking typecode
    add_newdoc('numpy._core.numerictypes', int_name, ('bit_count',
        f"""
        {int_name}.bit_count() -> int

        Computes the number of 1-bits in the absolute value of the input.
        Analogous to the builtin `int.bit_count` or ``popcount`` in C++.

        Examples
        --------
        >>> np.{int_name}(127).bit_count()
        7""" +
        (f"""
        >>> np.{int_name}(-127).bit_count()
        7
        """ if dtype(int_name).char.islower() else "")))
